using DataFrames
using BisectPy, Statistics

"""指标归一化"""
function normal(df::DataFrame)::DataFrame
    m, n = size(df)
    result = Matrix{Float64}(undef, m, n)
    for j in 1:n
        col_max = maximum(df[!, j])
        col_min = minimum(df[!, j])
        result[!, j] = (df[i, j] .- col_min) ./ (col_max - col_min)
    end
    return DataFrame(result, :auto)
end


"指标平均归一化"
function ave_normal(df::DataFrame)::DataFrame
    m, n = size(df)
    result = Matrix{Float64}(undef, m, n)
    for j in 1:n
        col_max = maximum(df[!, j])
        col_min = minimum(df[!, j])
        col_mean = mean(df[i, j])
        result[!, j] = (df[i, j] .- col_mean) ./ (col_max - col_min)
    end
    return DataFrame(result, :auto)
end


"""指标标准化"""
function standard(df::DataFrame)::DataFrame
    m, n = size(df)
    result = Matrix{Float64}(undef, m, n)
    for j in 1:n
        col_mean = mean(df[!, j])
        col_std = std(df[!, j])
        result[!, j] = (df[!, j] .- col_mean) ./ col_std
    end
    return DataFrame(result, :auto)
end


"""将正向指标利用模长缩放到单位长度"""
function scaling(df::DataFrame)::DataFrame
    m, n = size(df)
    result = Matrix{Float64}(undef, m, n)
    for j in 1:n
        model_length = sum(df[!, j] .^ 2)
        result[!, j] = df[!, j] ./ sqrt(model_length)
    end
    return DataFrame(result, :auto)
end


"""计算变异性指标"""
function variability(df::DataFrame)::Vector
    df = normalization(df)
    m, n = size(df)
    variabilities = []
    for j in 1:n
        ave = mean(df[!, j])
        sum_var = 0
        for i in 1:m
            sum_var += (df[i, j] - ave)^2 / (m - 1)
        end
        append!(variabilities, sqrt(sum_var))
    end
    return variabilities
end


"""计算相关系数矩阵"""
function correlation(df::DataFrame)::DataFrame
    df = normalization(df)
    _, n = size(df)
    result = Matrix{Float64}(undef, n, n)
    for j in 1:n
        for i in 1:n
            result[i, j] = result[j, i] = cor(df[!, i], df[!, j])
        end
    end
    return DataFrame(result, :auto)
end


"""计算冲突性指标"""
function conflict(df::DataFrame)::Vector
    corr_matrix = correlation(df)
    conflicts = []
    m, n = size(corr_matrix)
    for j in 1:n
        conflict = 0
        for i in 1:m
            conflict += (1 - corr_matrix[i, j])
        end
        append!(conflicts, conflict)
    end
    return conflicts
end


"""计算critic权重"""
function critic(df::DataFrame)::Vector
    information = variability(df) .* conflict(df)
    weight = information ./ sum(information)
    return weight
end


"""计算距离矩阵"""
function dist_matrix(df::DataFrame, bv_list::Vector)::DataFrame
    df = Matrix(df)
    m, n = size(df)
    if n != length(bv_list)
        error("输入的变量个数与矩阵的列数不符")
    else
        bv_matrix = repeat(bv_list', m)
    end
    dist_matrix = abs.(df - bv_matrix)
    return DataFrame(dist_matrix, :auto)
end


"""计算rsr得分"""
function rsr(df::DataFrame, bv_list::Vector)::Vector
    m, n = size(df)
    distance_matrix = dist_matrix(df, bv_list)
    rsr_matrix = Matrix{Float64}(undef, m, n)
    for j in 1:n
        compare_list = sort(distance_matrix[:, j])
        for i in 1:m
            rsr_matrix[i, j] = bisect_left(compare_list, distance_matrix[i, j])
        end
    end
    rsr_matrix
end


"""计算topsis得分"""
function topsis(df::DataFrame; weight=critic)::Vector
    df = scaling(df)
    weights = weight(df)
    m, n = size(df)
    result = Vector{Float64}(undef, m)
    D1 = D2 = Matrix{Float64}(undef, m, n)
    for j in 1:n
        col_max = maximum(df[!, j])
        col_min = minimum(df[!, j])
        for i in 1:m
            D1[i, j] = weights[j] * (df[i, j] - col_max)^2
            D2[i, j] = weights[j] * (df[i, j] - col_min)^2
        end
    end

    for i in 1:m
        result[i] = sqrt(sum(D1[i, :])) / (sqrt(sum(D1[i, :])) + sqrt(sum(D2[i, :])))
    end
    return result .* 100
end